import json

from . import utils 

class FileWriter():
    DELIMITER = '\n'

    def __init__(self, target_name):
        self.target_name = target_name
        self.target = None
        self.target_template = None
        self.first_written = False

    def _output_value(self, entry):
        return utils.force_string(entry.value)

    async def stream(self, source):
        async for entry in source:
            if '{input}' in self.target_name:
                self.target_template = self.target_name
            else:
                self.target = utils.get_file_io(self.target_name, write=True)

            output = self._output_value(entry)
            if self.first_written:
                output = self.DELIMITER + output
            else:
                self.first_written = True

            if self.target_template:
                with open(self.target_template.format(input=entry.original_value), 'w') as target_file:
                    target_file.write(output)
            else:
                self.target.write(output)
            if hasattr(self.target, 'flush'):
                self.target.flush()

        if self.target and hasattr(self.target, 'close'):
            self.target.close()

CONSUMERS = {
    'file': FileWriter,
}

def load_consumer(path):
    if path is None:
        return None
    elif '.' in path:
        Consumer = utils.import_obj(path)
    else:
        Consumer = CONSUMERS.get(path)
    return Consumer




from importlib import import_module
import json
import sys
import os

def get_file_io(name, write=False):
    # Test for file-like objects we can use first
    if hasattr(name, 'write') and write:
        return name
    elif hasattr(name, 'read') and not write:
        return name

    # Test for stdin/stdout special case
    if name == '-' and write:
        return sys.stdout
    elif name == '-' and not write:
        return sys.stdin

    # Assume this is a file
    return open(name, 'w' if write else 'r', 1)

def import_obj(path):
    # Ensure the current directory is on the path
    current_dir = os.path.abspath('.')
    if current_dir not in sys.path:
        sys.path.append(current_dir)

    # Split the object name and the module path
    module_path, handler_name = path.rsplit('.', 1)
    module = import_module(module_path)
    return getattr(module, handler_name)
    for attr in module_path.split('.')[1:]:
        module = getattr(module, attr)
    handler_obj = getattr(module, handler_name)
    if handler_obj is None:
        raise ValueError('Unable to import object: {}'.format(path))
    return handler_obj

def inject_module(module_name, namespace):
    try:
        module = import_module(module_name)
    except ModuleNotFoundError as e:
        print('This plugin requires module: "{}"\nTry "pip install {}"'.format(module_name, module_name))
        sys.exit(6)
    sys.modules[module_name] = module
    namespace[module_name] = module

def truthy(entries):
    return [entry for entry in entries if entry]

def arg_help(description, example=None):
    def decorator(streamer):
        if description is not None:
            streamer._arg_description = description
        if example is not None:
            streamer._arg_example = example
        return streamer
    return decorator

def force_string(value):
    if isinstance(value, str):
        return value
    try:
        value = json.dumps(value)
    except Exception as e:
        value = str(value)
    return value



import asyncio

async def drain(generator):
    """ A no-op drain of a generator """
    items = []
    async for item in generator:
        items.append(item)
    return items

async def transync(synchronous_iter):
    """
        This is a converter which takes a synchronous iterator and yields the items
        asynchronously which allows a standard pipe to convert to an async pipe.
    """
    for item in synchronous_iter:
        yield item

async def static_pipe(stream, inputs):
    source = transync(inputs)
    outputs = await drain(stream(source))
    return outputs

async def pipe(generator, streamers, consumer=None):
    """
        The "pipe" function is the central piece of a stream, the pipe connects
        the generators together in a chain allowing each to pass the result of the
        previous to the next.

        The first generator is special as it does not take an input stream. 

        The last piece of the stream is the consumer which is not a generator but 
        is expected to drain the generator given to it, writing any valued output.
    """
    if consumer is None:
        consumer = drain

    pipe = generator
    for streamer in streamers:
        pipe = streamer(pipe)

    await consumer(pipe)

def sync_exec(future):
    loop = asyncio.get_event_loop()
    task = asyncio.ensure_future(future, loop=loop)
    result = loop.run_until_complete(task)
    return result



from . import utils 
from . import entries

class FileReader():
    DELIMITER = '\n'

    def __init__(self, source_name, force_flush=False):
        self.source_name = source_name
        self.force_flush = force_flush

        self.first_written = False
        self.source = None

    async def stream(self):
        factory = entries.EntryFactory()
        source = utils.get_file_io(self.source_name)
        for index, line in enumerate(source):
            yield factory(line.rstrip(self.DELIMITER))

        if hasattr(source, 'close'):
            source.close()

GENERATORS = {
    'file': FileReader,
}

def load_generator(path):
    if path is None:
        return None
    elif '.' in path:
        Generator = utils.import_obj(path)
    else:
        Generator = GENERATORS.get(path)
    return Generator

